import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.awt.Point;

public class CourierHeuristic implements Heuristic {
   public CourierHeuristic(Set<Job> allJobs) {
      this.allJobs = allJobs;
      initialiseLookup();
   }

   /**
    * We want Manhattan distance
    */
   public int distance(Point a, Point b) {
      return Math.abs(a.x - b.x) + Math.abs(a.y - b.y);
   }

   /**
    * The magic happens here!
    */
   public <T extends Job> int estimate(Set<T> completed, Point end) {
      Set<Job> todo = new HashSet<Job>(allJobs);
      todo.removeAll(completed);

      int cost = 0;
      
      //get the travel distance for the remaining jobs
      for (Job j : todo) {
         cost += distance(j.start(), j.end());
      }

      //work out the minimum distance between the end of
      //each job and the start of ANY other
      //then we drop the largest to ensure it is an
      //underestimate (we are supposed to END at the
      //end of the last job)
      if (todo.size() > 1) {
         assert(lookup.containsKey(todo));
         cost += lookup.get(todo);
      } else if (todo.size() == 1) {
         cost += distance(end, todo.iterator().next().start());
      }

      return cost;
   }

   private void initialiseLookup() {
      lookup = new HashMap<Set<Job>, Integer>();
      int numJobs = allJobs.size();

      if (numJobs > 1) {
         List<Job> l = new ArrayList<Job>(allJobs);

         //1 << size is fast power 2 calculation for ints
         //remove the individual sets; we have a better heuristic here
         //and ignore the empty set -- its heuristic is zero!
         int len = 1 << l.size(); // - l.size() - 1;
         //System.out.println("we have " + numJobs + " nodes --> " + len + " sets");

         List<Set<Job>> powerset = new ArrayList<Set<Job>>(len);
         for (int field = 0; field < len; field++) {
            //get the bitfield in a form Java will let us use
            String bits = Integer.toBinaryString(field);
            while (bits.length() < numJobs) {
               bits = "0" + bits; //pad to the right length
            }

            powerset.add(new HashSet<Job>()); //initialise the entry

            //now treat the index as a bitfield
            //if bit for corresponding element is set, then add!
            for (int item = 0; item < numJobs; item++) {
               if (bits.charAt(item) == '1') {
                  powerset.get(field).add(l.get(item));
               }
            }
         }
   
         //we have the subsets, now compute their heuristics
         for (Set<Job> s : powerset) {
            if (s.size() > 1) {
               lookup.put(s, computeSpecificHeuristic(s));
            }
         }

      }
      //System.out.println("LOOKUP GENERATION COMPLETE! LET'S GO!");
   }

   private int computeSpecificHeuristic(Set<Job> set) {
      assert(set.size() > 1);
      int cost = 0;
      int d;
      int min = -1;
      int largest = 0;
      for (Job a : set) {
         for (Job b : set) {
            if (a != b) {
               d = distance(a.end(), b.start());
               if (d < min || min < 0) {
                  min = d;
               }
            }
         }
         cost += Math.max(0, min);
         if (min > largest) {
            largest = min;
         }
      }
      return cost - largest;
   }


   private Set<Job> allJobs;
   private Map<Set<Job>, Integer> lookup;
}

